#!/ops/tools/bin/perl
#
##  Copyright (c) 1995-2007 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------    
# @file       ChampTools.pm
# Routines for processing CHAMP data
# 
# @author     Doug Hunt
# @since      11/21/2007
# @version    $URL: svn://svn1.cosmic.ucar.edu/trunk/src/MOOutils/ChampParser.pm $ $Id: ChampParser.pm 17078 2014-02-03 20:20:38Z repro $
# cdaacTask   no
# -----------------------------------------------------------------------*/
EOD

package ChampParser;
use base qw(BJParser); # inherit from BJParser.pm

require 5.010_000;  # Needs features from perl version 5.10

use BJTools;
use TimeClass;
use vars qw($bigendian $tmpdir);
use Config;
use PDL;
use Carp;
use BJConstants;

# Global variables used to save time
my %dt_global;
my %zeroes_global_ushort;
my %zeroes_global_short;
my %ones_global;
my $hrcnt = 0;

#/**----------------------------------------------------------------------
# @sub       parse_OBSDqfit
#
# Subroutine to parse the main observable record:  OBSDqfit
# (for 'observation, quadratic fit')
# Input is the packet, with 'OBSDqfit' removed and the CRC removed
# (this routine is only called if the CRC is good).
#
# Output is a set of ASCII files in $$tmpdir (global) with low rate
# data records per antenna.  Also output is high rate data, per antenna,
# in binary files ready for conversion into opnGps format.
#
# @parameter  $self -- BJParser object
# @           $pkt -- Contents of OBSDqfit packet (without 'OBSDqfit')
# @                   as a perl string
# @return     $$tmpdir/lorate.txtXX, where XX is the antenna ID
# @           $$tmpdir/hirate.binXX, where XX is the antenna ID
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub parse_OBSDqfit {

  my $self = shift;
  my $pkt  = shift;

  my $bjdlength = length($pkt) + 8;  # length should include library ID and packet ID ('OBSDqfit')

  my $time = BJParser::parseStrip ("L", \$pkt); # 4

  my $date = '';
  if (defined($self->{BYDATE})) {
    # Inlined TimeClass for speed.
    my @mkta = gmtime($time + $TimeClass::GPSSEC);
    $date = sprintf ".%04d.%03d", $mkta[5] + 1900, $mkta[7] + 1;
  }
  my $tmpdir = $self->{TMPDIR};
  my $lv0name = defined($self->{ADDNAME}) ? '.'.Dplib::simplefn($self->{INFILE}) : '';

  my ($prn, $antin, $obstype, $sampint,
      $CA_chan, $CA_snr, $CA_faz, $CA_tau,
      $P2_chan, $P2_snr, $P2_faz, $P2_tau,
      $P1_chan, $P1_snr, $P1_faz, $P1_tau) = BJParser::parseStrip ("CCCC CSdd CSff CSff", \$pkt); # 45

  # print "OBSDqfit: $time, $prn, $antin\n";

  # Track status is always 0 for CHAMP
  my $trkstatus = 0;

  # Decode ObsType
  my $flywheel = $obstype & 0x20;
  my $l2c      = $obstype & 0x08; # Match the old code in BJfmtl_cosmic.c.  CHAMP should never have L2C data. D. Hunt 1/16/2014
  $flywheel = 1 if ($flywheel);
  $l2c      = 2 if ($l2c);

  # Convert the pseudoranges to meters
  my $Pcodechip = 299792458/10230000;  # P1 wavelength ~ 30m
  $CA_tau *= (10*$Pcodechip);
  $P1_tau *= $Pcodechip;
  $P2_tau *= $Pcodechip;

  # Observable data format (all on one line):
  # time (sec past J2000), OBSD, prn (prnxx), antenna index (hex),
  # obstype (hex--does not have stable definition--I currently ignore this one),
  # sample interval (seconds, hex), BlackJack packet size (bytes, decimal),
  # CA:, CA channel (hex), CA SNR, CA phase (cyc), CA pseudorange (meters),
  # P2:, (similar fields), P1:, (similar fields).
  if (length($pkt) == 0) {
    my $file = "lorate.txt$antin$date$lv0name";
    my $outrec_lr =
      sprintf ("%ld OBSD prn%02d %02x %02x %02x %d CA: %02x %d %lf %lf  P2: %02x %d %f %f  P1: %02x %d %f %f %2d %1d\n",
                       $time, $prn,
                       $antin, $obstype, $sampint, $bjdlength,
                       $CA_chan, $CA_snr, $CA_faz , $CA_tau,
                       $P2_chan, $P2_snr, ($CA_faz - $P2_faz)*$L1_L2, $CA_tau - $P2_tau,
                       $P1_chan, $P1_snr, $CA_faz - $P1_faz, $CA_tau - $P1_tau, $trkstatus, $flywheel);

    if ($tmpdir eq 'memory') {
      $self->{OUTFILES}{$file} .= $outrec_lr;
    } else {

      # Only print out low rate data to ASCII file.
      open my $lrfh, '>>', "$$tmpdir/lorate.txt$antin$date" or croak "Cannot open lorate.txt$antin$date for appending";
      printf {$lrfh} $outrec_lr;
      close $lrfh;
    }
      return;
  }

  #
  ## Now handle high rate data!
  #
  my ($CA_dop, $CA_ddop) = BJParser::parseStrip ("ff", \$pkt); # 8

  my %resid = ();

  # Rtyp:  For phase: CA 0xfc, P1 0xf1, P2 0xf2.  For amp: CA 0xac, P1 0xa1, P2 0xa2.
  my %dtype = (0xfc, 'CA_faz', 0xf1, 'P1_faz', 0xf2, 'P2_faz', 0xac, 'CA_amp', 0xa1, 'P1_amp', 0xa2, 'P2_amp');

 RESIDUAL:
  while (length($pkt) > 0) {

    my ($Rtyp, $Rscale, $Rrate) = BJParser::parseStrip ("Ccc", \$pkt); # 3

    # printf ("Rtyp, Rscale, Rrate, obstype = %x, $Rscale, $Rrate, %b\n", $Rtyp, $obstype);

    my $is_faz = ($Rtyp & 0x10);  # Phase measurements use signed residuals, amplitude uses unsigned residuals.  D. Hunt 1/17/2014
    my $dtyp = $dtype{$Rtyp} || do { carp "Unknown residual type: $Rtyp"; next RESIDUAL };

    # Read residuals into a hash of PDLs which is indexed by 'Rtyp'
    # Use memo-ized zeroes to save time.  Pre-declare separate zero vectors for each Rrate and dtyp
    $resid{$dtyp} = $is_faz ? $zeroes_global_short{$Rrate}{$dtyp}  //= zeroes(short,$Rrate)
                  :           $zeroes_global_ushort{$Rrate}{$dtyp} //= zeroes(ushort,$Rrate);

    $resid{$dtyp}->make_physical;
    ${$resid{$dtyp}->get_dataref} = substr ($pkt, 0, $Rrate*2, '');
    $resid{$dtyp}->upd_data;
    $resid{$dtyp}->bswap2 if (!$bigendian);
  }

  return if (keys %resid == 0); # no high rate resid

  #
  ## Compute high rate data vectors
  #
  my $Rrate = $resid{'CA_faz'}->nelem;

  #typedef struct opnGpsHdr {
  #  unsigned short rate
  #  unsigned short PRN
  #  unsigned int   GPSsec
  #  double         CArange
  #  unsigned char  datalen
  #  unsigned char  antid
  #  unsigned char  trkstatus
  #  unsigned char  flystatus
  #} __attribute__ ((packed)) opnGpsHdr
  my $hrhdr = pack ("SSIdCCCC", $Rrate, $prn, $time, $CA_tau, 0, $antin, $trkstatus, $flywheel | $l2c);

  # Use a memo-ized $dt (saves lots of time)
  my $dt = $dt_global{$Rrate} //= (sequence($Rrate)/$Rrate) - 0.5;

  # Attempt to correct 'wrapping'--bad compression when there is not enough dynamic range for residuals
  my $ca_resid = PDL::fixwrap($dt, $resid{'CA_faz'});
  my $p2_resid = PDL::fixwrap($dt, $resid{'P2_faz'});

  # debug
  # open DBG, ">>/tmp/champ_residuals.$prn";
  # wcols $time+$dt, $p2_resid, *DBG;
  # close DBG;

  # High rate values for CA faz and P2 faz
  my $ca = (( $CA_faz           + $CA_dop*$dt + $CA_ddop*$dt**2/2)           + $ca_resid/8192) * $LAMB1;
  my $p2 = ((($CA_faz - $P2_faz + $CA_dop*$dt + $CA_ddop*$dt**2/2) * $L1_L2) + $p2_resid/8192) * $LAMB2;

  # High rate CA SNR
  my $casnr = $SNRCONV50 * $resid{'CA_amp'} * 10;

  # Use memo-ized ones to save time
  my $p2snr = $P2_snr * 10 * ($ones_global{$Rrate} //= ones($Rrate));  # constant for entire second

  #
  ## Reformat these high rate values into an opnGps record
  #

  # typedef struct opnGpsHR {
  #   float          dtime
  #   double         l1faz
  #   double         l2faz
  #   unsigned short snr1
  #   unsigned short snr2
  #   double         camdl
  #   float          dfaz
  # } __attribute__ ((packed)) opnGpsHR
  #
  # typedef struct opnGps {
  #   opnGpsHdr      hdr
  #   opnGpsHR       hr[MAXHR]

  # } __attribute__ ((packed)) opnGps
  # Doing this with multiple appends and a 'reshape' is quicker than using 'cat' (which looks prettier)
  my $hrvals = $dt->append($ca)->append($p2)->append($casnr)->append($p2snr)->reshape($Rrate, 5);
  my $hrrec  = pack ("fddSS" x $Rrate, $hrvals->xchg(0,1)->flat->list);
  $hrcnt++;

  $file = "hirate.bin$antin$date$lv0name";
  if ($tmpdir eq 'memory') {
    $self->{OUTFILES}{$file} .= $hrhdr;
    $self->{OUTFILES}{$file} .= $hrrec;
  } else {

    # Append record onto output file
    open my $hrfh, '>>', "$$tmpdir/hirate.bin$antin$date" or croak "Cannot open hirate.bin$antin$date for appending";
    print {$hrfh} $hrhdr, $hrrec;
    close $hrfh;
  }

    return $time;

}

1;
