#!/ops/tools/bin/perl
#
##  Copyright (c) 1995-2007 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------    
# @file       CnofsParser.pm
# Routines for processing CNOFS data
# 
# @author     Doug Hunt
# @since      11/21/2007
# @version    $URL: svn://svn1.cosmic.ucar.edu/trunk/src/MOOutils/CnofsParser.pm $ $Id: CnofsParser.pm 15133 2012-09-12 16:59:46Z cosmicops $
# cdaacTask   no
# -----------------------------------------------------------------------*/
EOD

package CnofsParser;
use base qw(BJParser); # inherit from BJParser.pm

require 5.010_000;  # Needs features from perl version 5.10

use strict;
use warnings;
use BJTools;
use TimeClass;
use vars qw($bigendian $tmpdir);
use Config;
use PDL;
use PDL::NiceSlice;
use PDL::GSL::INTERP;
use Carp;
use BJConstants;


#/**----------------------------------------------------------------------
# @sub       parse_OBSDqfit
#
# Subroutine to parse the main observable record:  OBSDqfit
# (for 'observation, quadratic fit')
# Input is the packet, with 'OBSDqfit' removed and the CRC removed
# (this routine is only called if the CRC is good).
#
# Output is a set of ASCII files in $$tmpdir (global) with low rate
# data records per antenna.  Also output is high rate data, per antenna,
# in binary files ready for conversion into opnGps format.
#
# @parameter  $self -- BJParser object
# @           $pkt -- Contents of OBSDqfit packet (without 'OBSDqfit')
# @                   as a perl string
# @return     $$tmpdir/lorate.txtXX, where XX is the antenna ID
# @           $$tmpdir/hirate.binXX, where XX is the antenna ID
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub parse_OBSDqfit {

  my $self = shift;
  my $pkt  = shift;

  my $bjdlength = length($pkt) + 8;  # length should include library ID and packet ID ('OBSDqfit')

  my $time = BJParser::parseStrip ("L", \$pkt); # 4

  my $date = defined($self->{BYDATE}) ? '.'.TimeClass->new->set_gps($time)->get_yrdoy_gps : '';
  my $lv0name = defined($self->{ADDNAME}) ? '.'.$self->{INFILE} : '';

  my $tmpdir = $self->{TMPDIR};

  my ($prn, $antin, $obstype, $sampint,
      $CA_chan, $CA_snr, $CA_faz, $CA_tau,
      $P2_chan, $P2_snr, $P2_faz, $P2_tau) = BJParser::parseStrip ("CCCC CSdd CSff", \$pkt); # 34

  # print "OBSDqfit: $time, $prn, $antin\n";

  # Get Track Status from antenna index byte
  my $trkstatus = ($antin >> 2);

  # Get rid of any track status information which may be stored in the upper 6 bits
  # $antin = $antin & 0x03;
  $antin = 0;  # Not sure what $antin means, there is only one antenna for C/NOFS

  # pktlen ObsType
  # 260    00000111
  # 260    10000111
  # 260    10010111
  #  46    00001111
  #  46    10000011
  #  46    10001111
  # 466    00000111
  # 466    10000111
  my $flywheel  = $obstype & 0x20;

  # Convert the pseudoranges to meters
  my $Pcodechip = 299792458/10230000;  # P1 wavelength ~ 30m
  $CA_tau *= (10*$Pcodechip);
  $P2_tau *= $Pcodechip;

  # Observable data format (all on one line):
  # time (gps seconds), OBSD, prn (prnxx), antenna index (hex),
  # obstype (hex--does not have stable definition--I currently ignore this one),
  # sample interval (seconds, hex), BlackJack packet size (bytes, decimal),
  # CA:, CA channel (hex), CA SNR, CA phase (cyc), CA pseudorange (meters),
  # P2:, (similar fields), P1:, (similar fields).
  my $file = "lorate.txt$antin$date$lv0name";
  my $outrec_lr =
    sprintf ("%ld OBSD prn%02d %02x %02x %02x %d CA: %02x %d %lf %lf  P2: %02x %d %f %f  P1: %02x %d %f %f %2d %1d\n",
             $time, $prn,
             $antin, $obstype, $sampint, $bjdlength,
             $CA_chan, $CA_snr, $CA_faz , $CA_tau,
             $P2_chan, $P2_snr, ($CA_faz - $P2_faz)*$L1_L2, $CA_tau - $P2_tau,
             $CA_chan, $CA_snr, $CA_faz, $CA_tau, $trkstatus, $flywheel); # fill in CA for P1, P1 data are missing for C/NOFS!
  if ($tmpdir eq 'memory') {
    $self->{OUTFILES}{$file} .= $outrec_lr;
  } else {
    open my $lrfh, '>>', "$$tmpdir/$file" or croak "Cannot open $file for appending";
    printf {$lrfh} $outrec_lr;
    close $lrfh;
  }
  return if (length($pkt) == 0);
  return if ($sampint != 1); # High rate data only found for sample interval == 1

  #
  ## Now handle high rate data!
  #
  my ($CA_dop, $CA_ddop) = BJParser::parseStrip ("ff", \$pkt); # 8

  my %resid = ();

  # Rtyp:  For phase: CA 0xfc, P1 0xf1, P2 0xf2.  For amp: CA 0xac, P1 0xa1, P2 0xa2.
  my %dtype = (0xfc, 'CA_faz', 0xf1, 'P1_faz', 0xf2, 'P2_faz', 0xac, 'CA_amp', 0xa1, 'P1_amp', 0xa2, 'P2_amp');

 RESIDUAL:
  while (length($pkt) > 0) {

    my ($Rtyp, $Rscale, $Rrate) = BJParser::parseStrip ("Ccc", \$pkt); # 3

    # printf ("Rtyp, Rscale, Rrate, obstype = %x, $Rscale, $Rrate, %b\n", $Rtyp, $obstype);

    my $dtyp = $dtype{$Rtyp} || do { carp "Unknown residual type: $Rtyp"; next RESIDUAL };

    # Read residuals into a hash of PDLs which is indexed by 'Rtyp'
    $resid{$dtyp} = ($dtyp =~ /amp/) ? zeroes(ushort,$Rrate) : zeroes(short,$Rrate);
    $resid{$dtyp}->make_physical;
    ${$resid{$dtyp}->get_dataref} = substr ($pkt, 0, $Rrate*2, '');
    $resid{$dtyp}->upd_data;
    $resid{$dtyp}->bswap2 if (!$bigendian);
  }

  return if (keys %resid == 0); # no high rate resid

  # printf "%3d %8b %8b ", $bjdlength, $obstype, $antin;
  # print  join (" ", sort keys %resid), "\n";

  # Do not fake P2 data if it is missing!  D. Hunt 2010-11-02
  # This means that all single-frequency high rate ionospheric data from C/NOFS is discarded here!
  # return if (!exists ($resid{'P2_faz'}));

  # These 'CA_faz P2_faz' data are needed for the reference link.  Keep them in.  D. Hunt 2010/11/12
  # return if (!exists ($resid{'CA_amp'})); # Further, get rid of 'CA_faz P2_faz' tracks.  D. Hunt 2010/11/10

  #
  ## Compute high rate data vectors
  #
  my $Rrate = $resid{'CA_faz'}->nelem;

  # return;

  # Compute status values from obstype and antin bytes
  $flywheel     = ($obstype & 0x20)            ? 2**0 : 0;
  my $l2c       =                                       0; # no L2C from C/NOFS (yet)
  my $l2_onesec = (!exists ($resid{'P2_faz'})) ? 2**2 : 0;
  # Flystatus:  High rate data status byte
  #                 00000XYZ, where X = 1 for L1 only high rate
  #                                 Y = 1 for L2C (always 0 for C/NOFS)
  #                                 Z = 1 for Flywheeling
  my $flystatus = $l2_onesec | $flywheel | $l2c; # High rate status byte

  #typedef struct opnGpsHdr {
  #  unsigned short rate
  #  unsigned short PRN
  #  unsigned int   GPSsec
  #  double         CArange
  #  unsigned char  datalen
  #  unsigned char  antid
  #  unsigned char  trkstatus
  #  unsigned char  flystatus
  #} __attribute__ ((packed)) opnGpsHdr
  my $reserved = 0;
  my $hrhdr = pack ("SSLdCCCC", $Rrate, $prn, $time, $CA_tau, $reserved, $antin, $trkstatus, $flystatus);
  my $dt = (sequence($Rrate)/$Rrate) - 0.5; # ((double)i)/((double)Rrate) - 0.5;

  # Attempt to correct 'wrapping'--bad compression when there is not enough dynamic range for residuals
  my $ca_resid = PDL::fixwrap($dt, $resid{'CA_faz'});

  # High rate values for CA faz
  my $ca = (( $CA_faz + $CA_dop*$dt + $CA_ddop*$dt**2/2) + $ca_resid/8192) * $LAMB1;

  # If no HR P2_faz data, then write out only the one second values.  Otherwise, reconstruct
  # 50 Hz P2 phase normally
  my $p2;
  if (!$l2_onesec) { # full 50Hz L2 phase

    my $p2_resid = PDL::fixwrap($dt, $resid{'P2_faz'});
    $p2 = ((($CA_faz - $P2_faz + $CA_dop*$dt + $CA_ddop*$dt**2/2) * $L1_L2) + $p2_resid/8192) * $LAMB2;
  } else {

    # Set correct P2 phase value on even seconds ($dt == 0).  Set to -999 otherwise
    my $p2_middle_value = (($CA_faz - $P2_faz) * $L1_L2) * $LAMB2;
    $p2 = zeroes($dt->nelem) + $Dplib::err;
    $p2->where($dt == 0) .= $p2_middle_value;

    # For testing, leave commented in production code.  Try with zero residuals...
    # $p2 = ((($CA_faz - $P2_faz + $CA_dop*$dt + $CA_ddop*$dt**2/2) * $L1_L2) + 0/8192) * $LAMB2;
  }

  # High rate CA SNR
  my $casnr = exists ($resid{'CA_amp'}) ? $SNRCONV50 * $resid{'CA_amp'} * 10 : $CA_snr * 10 * ones($Rrate);
  my $p2snr = exists ($resid{'P2_amp'}) ? $SNRCONV50 * $resid{'P2_amp'} * 10 : $P2_snr * 10 * ones($Rrate);

  #
  ## Reformat these high rate values into an opnGps record
  #

  # typedef struct opnGpsHR {
  #   float          dtime
  #   double         l1faz
  #   double         l2faz
  #   unsigned short snr1
  #   unsigned short snr2
  #   double         camdl
  #   float          dfaz
  # } __attribute__ ((packed)) opnGpsHR
  #
  # typedef struct opnGps {
  #   opnGpsHdr      hdr
  #   opnGpsHR       hr[MAXHR]

  # } __attribute__ ((packed)) opnGps
  my $hrvals = cat ($dt, $ca, $p2, $casnr, $p2snr);
  my $hrrec  = pack ("fddSS" x $Rrate, $hrvals->xchg(0,1)->flat->list);

  $file = "hirate.bin$antin$date$lv0name";
  if ($tmpdir eq 'memory') {
    $self->{OUTFILES}{$file} .= $hrhdr;
    $self->{OUTFILES}{$file} .= $hrrec;
  } else {
    # Append record onto output file
    open my $hrfh, '>>', "$$tmpdir/$file" or croak "Cannot open $file for appending";
    print {$hrfh} $hrhdr, $hrrec;
    close $hrfh;
  }

  return $time;

}


#/**----------------------------------------------------------------------
# @sub       fill_onesec_l2
#
# Look for records in the input unsorted opnGps file which have only 1Hz L2 phase.
# Split these data into separate occultations and spline interpolate to 50Hz.
# Replace all into an output opnGps format file (still unsorted) without headers.
#
# @parameter  $infile    Input file in opnGps format, but out of order and
# @                      with no headers/trailers.
# @           $outfile   Output file with one second L2 phase data interpolated to 50Hz.
# @                      Still no headers/trailers.
# @return     none
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub fill_onesec_l2 {
  my $infile  = shift;
  my $outfile = shift;
  my $opt     = shift || {};

  # system "cp $infile $outfile"; # make a no-op! (for debug)
  # return;

  open my $infh, '<', $infile || die "Cannot open $infile";

  my %hr_data  = ();
  my %hdr_data = ();

  #
  ## Read in data and stash in a hash by PRN and time
  #
  my $hrlen  = 24;
  my $hdrlen = 20;

  my $overall_rate;

  open my $outfh, '>', $outfile || die "Cannot open $outfile for writing";

  while (1) {
    read ($infh, my $hdr = '', $hdrlen) or last;
    my ($rate, $prn, $t, $flystatus) = unpack ('SSL x[11] C', $hdr);
    $overall_rate //= $rate;
    die "Change in data rate:  Expecting $overall_rate, found $rate" if ($overall_rate != $rate);
    my $l2_onesec = ($flystatus & 0x04);
    read ($infh, my $hr = '', $rate * $hrlen) or last;

    # Put data with L2 phase to be interpolated into a hash.  Data which we can
    # leave alone are written directly to the output file
    if ($l2_onesec) {
      $hr_data{$prn}{$t}  = $hr;
      $hdr_data{$prn}{$t} = $hdr;
    } else {
      print {$outfh} $hdr . $hr;
    }
  }

  close ($infh);

  # Constant 50 Hz delta times
  my $dt = (sequence($overall_rate)/$overall_rate) - 0.5; # ((double)i)/((double)Rrate) - 0.5;

 PRN:
  foreach my $prn (1..32) {

    next PRN if (!exists($hr_data{$prn}));

    #
    ## Read all high rate data into 1D PDLs
    #
    my ($t_all, $p2_all) = (PDL->null, PDL->null);
    my @tsec = sort { $a <=> $b } keys %{$hr_data{$prn}}; # integer seconds
    my %one_sec_data = (); # Saving data that do not need to be interpolated for later writing back to output file

    foreach my $time (@tsec) {

      # Pull one second's worth of data (50 values) from packed binary records into 1D double PDLs
      my ($dt, $ca, $p2, $casnr, $p2snr) = pdl(unpack('fddSS' x $overall_rate, $hr_data{$prn}{$time}))->reshape(5, $overall_rate)->xchg(0,1)->dog;

      # Append this one second onto the whole time series for this PRN
      $t_all  = $t_all->append($time + $dt); # loss of precision?
      $p2_all = $p2_all->append($p2);

      # Save the rest of the one second data records for later repacking
      $one_sec_data{CA}{$time}    = $ca;
      $one_sec_data{CASNR}{$time} = $casnr;
      $one_sec_data{P2SNR}{$time} = $p2snr;
    }

    # Now break into separate occultation chunks
    my $del  = $t_all - $t_all->rotate(1);
    $del(0) .= 0;
    my $expected_interval = 1/$overall_rate; # 50Hz
    my $gaps = which(abs($del) > 2*$expected_interval);
    my $last_idx = 0;

    my $p2_filled = PDL->null; # Interpolated P2 phase for all occultations for this PRN

    # Loop over separate occultations (separated by gaps in time)
  GAP:
    foreach my $idx ($gaps->list, $t_all->nelem) {

      my $t_occ  = $t_all($last_idx:$idx-1);
      my $p2_occ = $p2_all($last_idx:$idx-1);

      # Get rid of -999's
      my $mask = ($p2_occ != $Dplib::err);
      my $t_input = $t_occ->where($mask);
      $p2_occ     = $p2_occ->where($mask);

      # ...and interpolate $p2_occ to 50Hz.  Disallow extrapolation for start and end 1/2 second beyond
      # the one second values.  These will be filled with -999 and discarded in read_opnGps.c (genXphase module)
      my $p2_out = PDL::GSL::INTERP->init('cspline', $t_input, $p2_occ)->eval($t_occ,
                                                                              {Extrapolate => 1});
      # my $p2_out = Dutil::dpspline ($t_input, $p2_occ, $t_occ, 9e30, $p2_occ->nelem);

      # Add this occultation back onto vector of all data for this PRN
      $p2_filled = $p2_filled->append($p2_out);

      $last_idx = $idx;
    }

    #
    ## Write records back with interpolated P2 included
    #
    my $reserved = 0;
    foreach my $tsec (@tsec) {
      my $idx    = which(($t_all >= $tsec-0.5) * ($t_all < $tsec+0.5));
      my $hrvals = cat ($dt, $one_sec_data{CA}{$tsec}, $p2_filled->index($idx), $one_sec_data{CASNR}{$tsec}, $one_sec_data{P2SNR}{$tsec});
      my $hrrec  = pack ("fddSS" x $overall_rate, $hrvals->xchg(0,1)->flat->list);
      print {$outfh} $hdr_data{$prn}{$tsec} . $hrrec;
    } # loop over integer seconds

  } # loop over PRNs

  close $outfh;

  return;
}
