#!/ops/tools/bin/perl
#
##  Copyright (c) 1995-2007 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------    
# @file       GraceTools.pm
# Routines for processing GRACE data
# 
# @author     Doug Hunt
# @since      11/21/2007
# @version    $URL: svn://svn1.cosmic.ucar.edu/trunk/src/MOOutils/GraceParser.pm $ $Id: GraceParser.pm 15133 2012-09-12 16:59:46Z cosmicops $
# cdaacTask   no
# -----------------------------------------------------------------------*/
EOD

package GraceParser;
use base qw(BJParser); # inherit from BJParser.pm

require 5.010_000;  # Needs features from perl version 5.10

use BJTools;
use TimeClass;
use vars qw($bigendian $tmpdir);
use Config;
use PDL;
use Carp;
use BJConstants;


#/**----------------------------------------------------------------------
# @sub       parse_OBSDqfit
#
# Subroutine to parse the main observable record:  OBSDqfit
# (for 'observation, quadratic fit')
# Input is the packet, with 'OBSDqfit' removed and the CRC removed
# (this routine is only called if the CRC is good).
#
# Output is a set of ASCII files in $$tmpdir (global) with low rate
# data records per antenna.  Also output is high rate data, per antenna,
# in binary files ready for conversion into opnGps format.
#
# @parameter  $self -- BJParser object
# @           $pkt -- Contents of OBSDqfit packet (without 'OBSDqfit')
# @                   as a perl string
# @return     $$tmpdir/lorate.txtXX, where XX is the antenna ID
# @           $$tmpdir/hirate.binXX, where XX is the antenna ID
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub parse_OBSDqfit {

  my $self = shift;
  my $pkt  = shift;

  my $bjdlength = length($pkt) + 8;  # length should include library ID and packet ID ('OBSDqfit')

  my $time = BJParser::parseStrip ("L", \$pkt); # 4

  my $date = defined($self->{BYDATE}) ? '.'.TimeClass->new->set_gps($time)->get_yrdoy_gps : '';
  my $tmpdir = $self->{TMPDIR};

  my ($prn, $antin, $obstype, $sampint,
      $CA_chan, $CA_snr, $CA_faz, $CA_tau,
      $P2_chan, $P2_snr, $P2_faz, $P2_tau,
      $P1_chan, $P1_snr, $P1_faz, $P1_tau) = BJParser::parseStrip ("CCCC CSdd CSff CSff", \$pkt); # 45

  # print "OBSDqfit: $time, $prn, $antin\n";

  # Get Track Status from antenna index byte
  my $trkstatus = ($antin >> 2);

  # Get rid of any track status information which may be stored in the upper 6 bits
  $antin = $antin & 0x03;

  # Decode ObsType
  my $flywheel = $obstype & 0x20;

  # Convert the pseudoranges to meters
  my $Pcodechip = 299792458/10230000;  # P1 wavelength ~ 30m
  $CA_tau *= (10*$Pcodechip);
  $P1_tau *= $Pcodechip;
  $P2_tau *= $Pcodechip;

  # Observable data format (all on one line):
  # time (sec past J2000), OBSD, prn (prnxx), antenna index (hex),
  # obstype (hex--does not have stable definition--I currently ignore this one),
  # sample interval (seconds, hex), BlackJack packet size (bytes, decimal),
  # CA:, CA channel (hex), CA SNR, CA phase (cyc), CA pseudorange (meters),
  # P2:, (similar fields), P1:, (similar fields).
  open my $lrfh, '>>', "$$tmpdir/lorate.txt$antin$date" or croak "Cannot open lorate.txt$antin$date for appending";
  printf {$lrfh} ("%ld OBSD prn%02d %02x %02x %02x %d CA: %02x %d %lf %lf  P2: %02x %d %f %f  P1: %02x %d %f %f %2d %1d\n",
                  $time, $prn,
                  $antin, $obstype, $sampint, $bjdlength,
                  $CA_chan, $CA_snr, $CA_faz , $CA_tau,
                  $P2_chan, $P2_snr, ($CA_faz - $P2_faz)*$L1_L2, $CA_tau - $P2_tau,
                  $P1_chan, $P1_snr, $CA_faz - $P1_faz, $CA_tau - $P1_tau, $trkstatus, $flywheel);
  close $lrfh;
  return if (length($pkt) == 0);

  #
  ## Now handle high rate data!
  #
  my ($CA_dop, $CA_ddop) = BJParser::parseStrip ("ff", \$pkt); # 8

  my %resid = ();

  # Rtyp:  For phase: CA 0xfc, P1 0xf1, P2 0xf2.  For amp: CA 0xac, P1 0xa1, P2 0xa2.
  my %dtype = (0xfc, 'CA_faz', 0xf1, 'P1_faz', 0xf2, 'P2_faz', 0xac, 'CA_amp', 0xa1, 'P1_amp', 0xa2, 'P2_amp');

 RESIDUAL:
  while (length($pkt) > 0) {

    my ($Rtyp, $Rscale, $Rrate) = BJParser::parseStrip ("Ccc", \$pkt); # 3

    # printf ("Rtyp, Rscale, Rrate, obstype = %x, $Rscale, $Rrate, %b\n", $Rtyp, $obstype);

    my $dtyp = $dtype{$Rtyp} || do { carp "Unknown residual type: $Rtyp"; next RESIDUAL };

    # Read residuals into a hash of PDLs which is indexed by 'Rtyp'
    $resid{$dtyp} = ($Rtyp == 0xAC) ? zeroes(ushort,$Rrate) : zeroes(short,$Rrate);
    $resid{$dtyp}->make_physical;
    ${$resid{$dtyp}->get_dataref} = substr ($pkt, 0, $Rrate*2, '');
    $resid{$dtyp}->upd_data;
    $resid{$dtyp}->bswap2 if (!$bigendian);
  }

  return if (keys %resid == 0); # no high rate resid

  #
  ## Compute high rate data vectors
  #
  my $Rrate = $resid{'CA_faz'}->nelem;

  # If these are the odd 10 second data with one second residuals,
  # then leave them alone.
  return $time if ($Rrate == 10);

  #typedef struct opnGpsHdr {
  #  unsigned short rate
  #  unsigned short PRN
  #  unsigned int   GPSsec
  #  double         CArange
  #  unsigned char  datalen
  #  unsigned char  antid
  #  unsigned char  trkstatus
  #  unsigned char  flystatus
  #} __attribute__ ((packed)) opnGpsHdr
  my $hrhdr = pack ("SSIdCCCC", $Rrate, $prn, $time, $CA_tau, $bjdlength, $antin, $trkstatus, $flywheel);

  my $dt = (sequence($Rrate)/$Rrate) - 0.5; # ((double)i)/((double)Rrate) - 0.5;

  # Attempt to correct 'wrapping'--bad compression when there is not enough dynamic range for residuals
  my $ca_resid = PDL::fixwrap($dt, $resid{'CA_faz'});
  my $p2_resid = PDL::fixwrap($dt, $resid{'P2_faz'});

  # debug
  # open DBG, ">>/tmp/grace_residuals.$prn";
  # wcols $time+$dt, $p2_resid, *DBG;
  # close DBG;

  # High rate values for CA faz and P2 faz
  my $ca = (( $CA_faz           + $CA_dop*$dt + $CA_ddop*$dt**2/2)           + $ca_resid/8192) * $LAMB1;
  my $p2 = ((($CA_faz - $P2_faz + $CA_dop*$dt + $CA_ddop*$dt**2/2) * $L1_L2) + $p2_resid/8192) * $LAMB2;

  # High rate CA SNR
  my $casnr = $SNRCONV50 * $resid{'CA_amp'} * 10;
  my $p2snr = $P2_snr * 10 * ones($Rrate);  # constant for entire second

  #
  ## Reformat these high rate values into an opnGps record
  #

  # typedef struct opnGpsHR {
  #   float          dtime
  #   double         l1faz
  #   double         l2faz
  #   unsigned short snr1
  #   unsigned short snr2
  #   double         camdl
  #   float          dfaz
  # } __attribute__ ((packed)) opnGpsHR
  #
  # typedef struct opnGps {
  #   opnGpsHdr      hdr
  #   opnGpsHR       hr[MAXHR]

  # } __attribute__ ((packed)) opnGps
  my $hrvals = cat ($dt, $ca, $p2, $casnr, $p2snr);
  my $hrrec  = pack ("fddSS" x $Rrate, $hrvals->xchg(0,1)->flat->list);

  # Append record onto output file
  open my $hrfh, '>>', "$$tmpdir/hirate.bin$antin$date" or croak "Cannot open hirate.bin$antin$date for appending";
  print {$hrfh} $hrhdr, $hrrec;
  close $hrfh;

  return $time;

}

1;
