#
##  Copyright (c) 1995-2011 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------
# @file       Level1aTools.pm
#
# Tools for converting raw data into level 1a data formats.
#
# @author     Doug Hunt
# @since      10/07/2003
# @cdaacTask  no
# @version    $URL: svn://svn1.cosmic.ucar.edu/trunk/src/BJutils/Level1aTools.pm $ $Id: Level1aTools.pm 22447 2016-12-01 03:58:34Z dhunt $
# -----------------------------------------------------------------------*/
EOD

package Level1aTools;

use strict;
use warnings;
use FileHandle;
use PDL;
use BJConstants;
use File::Basename;
use File::Temp;


#/**----------------------------------------------------------------------    
# @sub       blackJack2rinex
#
# Convert an ASCII file in the ad-hoc BJfmtl format to RINEX 2.20 format
# similar to a podObs file.
#
# @parameter  $infile    Input ASCII file.
# @           $outfile   Output Hatenaka compressed, gzipped RINEX file.
# @return     none
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub blackJack2rinex {

  my $infile    = shift;
  my $outfile   = shift;
  my $antid     = shift;
  my $leoname   = shift;
  my $mission   = shift;
  my $interval  = shift || 0; # if no interval specified, do not restrict based on interval
  my $options   = shift || {compress => 1};

  my ($outmem, $rinex_body, $rinex_header);
  my ($fileyr, $filedoy) = ($outfile =~ /(\d\d\d\d)\.(\d\d\d)/);
  if ($options->{rinex_header}) {
    $rinex_body = $options->{rinex_body};
    $rinex_header = $options->{rinex_header};
  } else {
    $$rinex_body   = '';
    $$rinex_header = '';
  }
  my $filegps = TimeClass->new->set_yrdoyhms_gps($fileyr, $filedoy, 0, 0, 0)->get_gps;

  open my $infh,  '<', $infile  or die "Cannot open input file $infile";

  my $duplicate_count = 0; # count of duplicate records for a given time stamp
  my $endtime = 0;
  my %rec = ();  # hash by PRN of data for one time step.  Eash entry is another hash
                 # with all the data types shown below in %hash.
  my %lastRec = (); # hash with all data for the previous record.  Useful in determining
                    # which duplicate record to use if there is more than one for a given
                    # PRN.  The rule is that tracks should, when possible, continue to 
                    # use the same obstype.

  my @times = ();

 RECORD:
  while (<$infh>) {
    # use nifty hash slice for record
    my %hash = ();
    @hash{'time', 'obsd', 'prn', 'antidx', 'obstype', 'interval', 'pcktsize',
	  'ca', 'cachan', 'casnr', 'caphase', 'capseudo',
	  'p2', 'p2chan', 'p2snr', 'p2phase', 'p2pseudo',
	  'p1', 'p1chan', 'p1snr', 'p1phase', 'p1pseudo'} = split;

    next unless (defined ($hash{obsd}) && $hash{obsd} =~ /OBSD/);         # skip nav data
    next unless ($hash{time} =~ /\d+/);          # make sure time is numeric
    next unless (abs($filegps - $hash{time}) < 604800); # skip records with times way off expected time. D. Hunt 4/11/2007

    next if ($interval && (hex($hash{interval}) != $interval)); # reject records with data rate != to that requested D. Hunt 7/27/2004
                                                 # Needed to allow processing of medium rate CHAMP data.

    $hash{prn}    = substr($hash{prn},3,2);  # get rid of 'prn' in eg 'prn08'
    my $prn = 'G' . $hash{prn}; # Add G to PRN number
    $hash{cachan} = hex($hash{cachan});

    # check for out of order records
    if ($hash{time} < $endtime) {
      warn "Out of order record:  time = $hash{time}, prn = $hash{prn}";
      next RECORD;
    }

    # write binex record when all prns for this epoch gathered
    if ($endtime != $hash{time} && $endtime != 0) {
      push (@times, $endtime);
      $$rinex_body .= writeRinex (\%rec, $endtime);
      %lastRec = %rec; undef %rec;
    }

    # If this is a duplicate entry for this PRN, we decide if we should use this record
    # or the currently existing record.  We want to keep the obstype the same as the last record
    # if possible.
    if ($rec{$prn}) {

      # Comment these print statements out because they make too
      # much noise in the output files.  Now there is the possibility
      # of processing 'double dump' cosLv0 files (dump_new, dump_prev)
      # so there will be many of these messages.  D. Hunt 9/25/2006
      $duplicate_count++;

      # Don't add %hash to %rec if this would cause a break in obstypes
      next RECORD if (defined ($lastRec{$prn}) &&
                      ($lastRec{$prn}{obstype} ne $hash{obstype}));
    }

    # Hash of prns per time step.  Used to save obstypes from the last time step
    # in order to help decide which observations to use if there are two observations
    # for a given time step and PRN.
    $rec{$prn} = \%hash;
    $endtime = $hash{time};
  }

  # Write last record
  $$rinex_body .= writeRinex (\%rec, $endtime);
  push (@times, $endtime);

  close ($infh);

  my ($tstart, $tend) = @times[0,-1];

  return if ($tend == 0); # no records written, bail out

  if (exists ($options->{rinex_header})) {
    $$rinex_header = writeRinexHeader ($antid, $leoname, $tstart, $tend, $mission);
    return; # $$rinex_body is already filled up
  }

  my $outfh;
  if($options->{compress}){
    open $outfh, "| rnx2crx - -s -e 100 | gzip - > $outfile";
  }else{
    $outfile =~ s/_crx\.gz$//i;
    open $outfh, "> $outfile";
  }
  print  {$outfh} writeRinexHeader ($antid, $leoname, $tstart, $tend, $mission);
  print  {$outfh} $$rinex_body;
  close  ($outfh);

  print "Found $duplicate_count records for duplicate times\n" if ($duplicate_count);
  return;

}


#/**----------------------------------------------------------------------
# @sub       writeRinex
#
# Write one time step of RINEX data
#
# @parameter
# @           $rec      Record hashref by PRN with input data for one time step
# @return     $data     The RINEX text of the record
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub writeRinex {

  my %rec  = %{(shift)};
  my $time = shift;
  my $data = '';

  my @obs = qw(L1 L2 C1 P1 P2 LA SA S2);

  # From [IGSMAIL-3281]: RINEX Modifications for LEO Satellites (Vers. 2.20)
  # Werner Gurtner <gurtner @ aiub.unibe.ch>
  # Lou Estey   <lou @ unavco.ucar.edu>
  # April 12, 2001
  # L1, L2:  L1, L2 phase, derived from P code
  # P1, P2:  P1, P2 pseudorange from P code
  # C1:      Pseudorange using C/A-Code on L1
  # LA:      Phase measurements on L1 derived from C/A code tracking
  # SA:      Raw signal strength or SNR for LA
  # S2:      Raw signal strengths or SNR values as given by the receiver for the L2 phase observation

  my $dataformat  = "%14.3f"."%16.3f" x 4 ."\n"."%14.3f"."%16.3f" x (@obs - 6) ."\n";

  my @prnlist = sort keys %rec;

  # Inlined from TimeClass for speed
  my @mkta  = gmtime($time + $TimeClass::GPSSEC);
  my @epoch = (substr($mkta[5] + 1900, 2, 2), $mkta[4] + 1, $mkta[3], $mkta[2], $mkta[1], $mkta[0]);

  my $epochformat = " %0.2d"."%3d"    x 4 ."%11.7f"."  0%3d";
  if (@prnlist <= 12) {
    $epochformat .= "%3s" x @prnlist ."\n";
  } else { # Allow for continuation line in case of more than 12 PRNs.
    $epochformat .= "%3s" x 12 ."\n"." " x 32 ."%3s" x (@prnlist - 12) ."\n";
  }
  $data .= sprintf $epochformat, @epoch, scalar(@prnlist), @prnlist;
  foreach my $prn (@prnlist) {
    $data .= sprintf $dataformat, @{$rec{$prn}}{qw(p1phase p2phase capseudo p1pseudo p2pseudo caphase casnr p2snr)};
  }

  return $data;

}



#/**----------------------------------------------------------------------
# @sub       writeRinexHeader
#
# Write the header for a RINEX file
#
# @parameter
# @           $antid    Antenna ID (integer)
# @           $tstart   Start time of data in GPS seconds
# @           $tend     End time of data in GPS seconds
# @           $mission  Mission name
# @return     $data     The RINEX header text
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub writeRinexHeader {

  my $antid   = shift;
  my $leoname = shift;
  my $tstart  = shift;
  my $tend    = shift;
  my $mission = $ENV{cdaac_mission} = shift;
  my $data    = '';

  # Get program name and version.
  my $prgname = $0;
  my $longheader = `strings $prgname | grep -m 1 '\$Id:'` // '';
  ($prgname = basename($prgname)) =~ s/\.pl//;
  $prgname = $mission if (length($prgname) >= 13);
  my ($f, $version, $date, $time, $user) = ($longheader =~ m{\$Id:\s+([\w\.]+)\s+(\d+)\s+([\d\-]+)\s+([\d\:Z]+)\s+(\w+)});
  if (defined($version)) {
    $prgname .= " v$version";
  }
  die "PGM name ($prgname) and version\# too long" if (length($prgname) > 20);

  # Defaults for RINEX header if no other info.
  my $runby = "COSMIC CDAAC";
  my $marktype = "SPACEBORNE";
  my $rectype = "";
  my $anttype = "";
  my $recnum = "";
  my $antnum = "";
  my $recver = "";
  my $rnxver =  2.2;

  my @obs = qw(L1 L2 C1 P1 P2 LA SA S2);
  my @wavefac = (1,1);
  my @c_mass = (0,0,0);
  my @antdel = (0,0,0);
  my @antbor = (0,0,0);

  # Read Antenna parameters from the Config.pm file.
  my $config    = do "$mission/Config.pm";
  die "Cannot find Config.pm file for $mission" if (!defined($config));
  my %antennas  = %{$config->{'satellite'}{'antennas'}};
  my ($antname) = grep { $antid == $antennas{$_}{'id'} } keys %antennas;
  my $nleo      = ($leoname =~ /^CHAM/) ? 0
                : ($leoname =~ /^CNFS/) ? 0
                : ($leoname =~ /^C/)    ? (substr $leoname, 1)-1 # COSMIC: 1-6, convert to index: 0-5
                : ($leoname =~ /^GRC/)  ? (substr $leoname, 3)-1 # GRACE 1-2 (GRACE-A or GRACE-B), convert to index, 0-1
                : 0;
  @antdel       = ref($antennas{$antname}{'offset'}->[0]) eq 'ARRAY'
                ? @{$antennas{$antname}{'offset'}->[$nleo]}
                : @{$antennas{$antname}{'offset'}};
  @antbor       = ref($antennas{$antname}{'boresight'}->[0]) eq 'ARRAY'
                ? @{$antennas{$antname}{'boresight'}->[$nleo]}
                : @{$antennas{$antname}{'boresight'}};

  # Determine MARKER NAME.  Special logic for cosmic_GOX_reader
  # C006 L26 (POD2)                                             MARKER NAME
  my $bernid   = $config->{'satellite'}{'id_num'}[$nleo];
  my $markname = sprintf "$leoname L%02d ($antname)", $bernid-900;
  $rectype  = $config->{'satellite'}{'rectype'};
  $anttype  = $config->{'satellite'}{'antennas'}{$antname}{'fullname'};

  # OBSERVER / AGENCY
  my %obsagency = (gpsmet => ["GPS/MET", "UCAR"],
                   champ  => ["CHAMP", "GFZ"],
                   grace  => ["GRACE", "GFZ"],
                   tsx    => ["TSRX", "GFZ"],
                   tdx    => ["TDMX", "GFZ"],
                   metopa => ["MTPA", "EUMETSAT"],
                   metopb => ["MTPB", "EUMETSAT"],
                   sacc   => ["SAC-C", "CONAE/NASA"],
                   cosmic => ["FORMOSAT3/COSMIC", "NSPO/UCAR"],
                   coseq  => ["FORMOSAT7/COSMIC-2E", "NSPO/UCAR"],
                   cos2e  => ["FORMOSAT7/COSMIC-2E", "NSPO/UCAR"],
                   kompsat5 => ["KOMPSAT5", "KASI/KARI"],
                   cnofs  => ["C/NOFS", "US AIR FORCE"],
                );
  my $proto = $mission;
  die "Proto mission $proto not recognized" unless exists $obsagency{$proto};
  my $observer = $obsagency{$proto}->[0];
  my $agency = $obsagency{$proto}->[1];

  # Get times of first and last observation.
  my @firstepoch = TimeClass->new->set_gps($tstart)->get_ymdhms_gps;
  my @lastepoch  = TimeClass->new->set_gps($tend)->get_ymdhms_gps;

  # Get date and time of this processing.
  my $timestamp = TimeClass->new->now->get_bernestamp_gps;

  # Write RINEX header.
  $data .= sprintf "%9.2f"." " x 11 ."%-20s" x 2 ."%1s\n", $rnxver, "OBSERVATION DATA", "M (MIXED)", "RINEX VERSION / TYPE";
  $data .= sprintf "%-20s" x 3 ."%1s\n", $prgname, $runby, $timestamp, "PGM / RUN BY / DATE";
  $data .= sprintf "%-60s%1s\n", $markname, "MARKER NAME";
  $data .= sprintf "%-60s%1s\n", $marktype, "MARKER TYPE";
  $data .= sprintf "%-20s%-40s%1s\n", $observer, $agency, "OBSERVER / AGENCY";
  $data .= sprintf "%-20s" x 3 ."%1s\n", $recnum, $rectype, $recver, "REC \# / TYPE / VERS";
  $data .= sprintf "%-20s%-40s%1s\n", $antnum, $anttype, "ANT \# / TYPE";
  $data .= sprintf "%14.4f" x 3 ." " x 18 ."%1s\n", @antdel, "ANTENNA: DELTA X/Y/Z";
  $data .= sprintf "%14.4f" x 3 ." " x 18 ."%1s\n", @antbor, "ANTENNA: B.SIGHT XYZ";
  $data .= sprintf "%14.4f" x 3 ." " x 18 ."%1s\n", @c_mass, "CENTER OF MASS: XYZ";
  $data .= sprintf "%6d" x 2 ." " x 48 ."%1s\n", @wavefac, "WAVELENGTH FACT L1/2";
  $data .= sprintf "%6d" . "%6s" x @obs . " " x (54 - 6*@obs) ."%1s\n", scalar(@obs), @obs, "\# / TYPES OF OBSERV";
  $data .= sprintf "%6d" x 5 ."%13.7f" ." " x 5 ."%3s" ." " x 9 . "%1s\n", @firstepoch, "GPS", "TIME OF FIRST OBS";
  $data .= sprintf "%6d" x 5 ."%13.7f" ." " x 5 ."%3s" ." " x 9 . "%1s\n", @lastepoch, "GPS", "TIME OF LAST OBS";
  $data .= sprintf " " x 60 ."%1s\n", "END OF HEADER";

  return $data;

}

#/**----------------------------------------------------------------------
# @sub       sortLrObs
#
# Sort Low rate data ASCII files, as output by the GRACE parsing software
# and the BJfmtl* family of C codes from JPL.
#
# @parameter  $infile    Input file in opnGps format, but out of order and
# @                      with no headers/trailers.
# @           $outfile   Output sorted file.
# @return     none
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub sortLrObs {
  my $infile  = shift;
  my $outfile = shift;

  my %data = ();

  #
  ## Read in data and stash in a hash by PRN and time
  #
  open my $infh, '<', $infile or die "Cannot open $infile";
  while (my $line = <$infh>) {
    my ($time, $prn) = ($line =~ /^(\d+) OBSD prn(\d+)/);
    $data{$time}{$prn} = $line;
  }
  close $infh;

  #
  ## Print data to output file sorted by time.
  #
  open my $outfh, '>', $outfile or die "Cannot open $outfile for writing";
  foreach my $time (sort { $a <=> $b } keys %data) {
    foreach my $prn (sort { $a <=> $b } keys %{$data{$time}}) {
      print {$outfh} $data{$time}{$prn};
    }
  }
  close $outfh;

  return;
}

#/**----------------------------------------------------------------------
# @sub       sortOpnGps
#
# Sort high rate data into the correct order after BJfmtl_cosmic
# is called.  Add headers and trailers.
#
#
# @parameter  $infile    Input file in opnGps format, but out of order and
# @                      with no headers/trailers.
# @           $outfile   Output sorted file.
# @return     none
# @exception  Will toss an exception in case of real trouble.
# ----------------------------------------------------------------------*/
sub sortOpnGps {
  my $infile  = shift;
  my $outfile = shift;
  my $opt     = shift || {};

  my $version = defined($opt->{VERSION}) ? $opt->{VERSION} : 1; # default to open loop format

  my $accept_odd_rates = $opt->{ODD_RATES};

  open (IN, $infile) || die "Cannot open $infile";

  my %data = ();

  #
  ## Read in data and stash in a hash by PRN and time
  #
  my $hrlen = ($version == 1) ? 36 # open loop
            : ($version == 2) ? 24 # closed loop
            : ($version == 3) ? 48 # l2c open loop
            : 0;
  die "bad version number: $version" if ($hrlen == 0);

  my $hdrlen = ($version == 3) ? 28 # l2c open loop
             : 20;

  while (1) {
    read (IN, my $hdr = '', $hdrlen) or last;
    my ($rate, $prn, $t) = unpack ('SSL', $hdr);
    my $nok = $accept_odd_rates ? ($prn < 1 || $prn > 32 || $t < 1)
      :                           ($rate < 46 || $rate > 52 || $prn < 1 || $prn > 32 || $t < 1);
    die "Bad rate, prn or time: $rate, $prn, $t" if ($nok);
    read (IN, my $hr = '', $rate * $hrlen) or last;
    $data{$prn}{$t} = $hdr . $hr;
  }

  close (IN);

  #
  ## Print data to output file sorted by PRN and time.
  ## Keep track of byte offsets of beginning of each PRN section
  #
  my @offsets = (-1) x 33;
  my $byteCount = 0;
  open (OUT, ">$outfile") || die "Cannot open $outfile for writing";
  foreach my $prn (1..32) {

    $offsets[$prn] = (exists $data{$prn}) ? $byteCount : 0;

    foreach my $time (sort { $a <=> $b } keys %{$data{$prn}}) {
      print OUT $data{$prn}{$time};
      $byteCount += length($data{$prn}{$time});
    }
  }

  # Append PRN index block to end of opnGps file
  print OUT pack ("l" x 32, @offsets[1..32]);

  # Append Trailer block describing version and column layout

  # Version 1, high rate data in "fddSSdf" format, one second data in "SSLdCCCC" format
  # (see perldoc -f pack)
  if      ($version == 1) {
    print OUT pack ("Ca31a32", 1, "fddSSdf",   "SSLdCCCC");  # 64 bytes of trailer
  } elsif ($version == 2) { # closed loop data (no CAmodel or dfaz information)
    print OUT pack ("Ca31a32", 2, "fddSS",     "SSLdCCCC");  # 64 bytes of trailer
  } elsif ($version == 3) { # L2C open loop data (P2 range, P2 phase model and P2 delta phase included)
    print OUT pack ("Ca31a32", 3, "fddSSdfdf", "SSLdCCCCd"); # 64 bytes of trailer
  }
  close OUT;

  return;
}


1;

