#!/usr/bin/env perl
#
##  Copyright (c) 1995-2018 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------
# @file       champZero2one.pl
# Convert one level 0 CHAMP data file into level 1 files as specified
# in the COSMIC level 1 data standard:
#
# http://cdaac-www.cosmic.ucar.edu/cdaac/fileFormats/
#
# @author     Doug Hunt
# @since      2/1/2018
# @version    $URL$ $Id$
# @cdaacTask  yes
# @usage      champZero2one.pl chpLv0_2008.001.04.58_bj
# -----------------------------------------------------------------------*/
EOD

$|++;

# Perl pragmas
use lib qw (.);
use strict;
use warnings;

# Perl standard libraries
use File::Temp qw(tempdir);
use Getopt::Long;

# Perl extensions
use PDL;

# CDAAC libraries
use BJTools;
use ChampParser;
use Level1aTools;

if (@ARGV < 1) {
  print $pkgdoc;
  exit -1;
}

my $inpath  = shift;

# Find date, LEO ID and dump number from input path: grcLv0_2016.366.02.36.001_bj
my ($date, $hr, $min) = ($inpath =~ /chpLv0_(\d\d\d\d\.\d\d\d)\.(\d\d)\.(\d\d)_bj/);
my $ndump = 1;
my $leoid = 1;

# Create a temporary working directory.  This returns a ref to a scalar,
# the scalar contains the name of the temp dir.
my $tmpdir = tempdir( CLEANUP => 1 );
my $tmpdirref = \$tmpdir;

# start fresh!
unlink glob ("$tmpdir/*");

print "Processing $inpath...\n";

my $parser = ChampParser->new({TMPDIR => $tmpdirref});

my $filetext = do { local( @ARGV, $/ ) = $inpath; <> } ; # slurp!

# get rid of the champ file wrappers:  4 byte packet ID and lengths
my $unwrapped = champ_unwrap(\$filetext);

$parser->parseLv0($unwrapped);
$parser->report; # print report of parsing success

print "Extracting NAV data\n";
my $logdata = BJTools::logData ("$tmpdir/navg.txt");

# Print out any GOX log data
if (length($logdata) > 0) {
  my $logfile = sprintf ("goxSOH_$date.%03d.%02d_txt", $leoid, $ndump);
  open (OUT, ">$logfile") || die "Cannot open $logfile for output";
  print OUT $logdata;
  close OUT;
  print "FILE CREATION: $logfile\n";
}


my @lrfiles = glob ("$tmpdir/lorate*");

print "Converting ASCII low rate data to level 1a format (RINEX)\n";
foreach my $lrtmp (@lrfiles) {

  # sort by time
  Level1aTools::sortLrObs($lrtmp, $lrtmp);

  my ($antenna_id) = $lrtmp =~ /lorate\.txt(\d+)/;

  #
  ## Write out 10 second data in compressed RINEX format
  #
  my $outfile = sprintf ("podCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  my $leoname = 'CHAM';
  my $mission = 'champ';
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 10, {compress => 0});
  unlink $outfile if (-s $outfile && -s $outfile < 2000);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

  #
  ## Write out 1 second data in compressed RINEX format
  #
  $outfile = sprintf ("ionCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 1, {compress => 0});
  unlink $outfile if (-s $outfile && -s $outfile < 2000);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

}

print "Converting high rate data to opnGps format\n";
my @if  = glob ("$tmpdir/hirate*");
foreach my $if (sort @if) {

  my ($antenna_id) = $if =~ /hirate\.bin(\d+)/;

  my $outfile = sprintf ("opnGps_$date.%03d.%02d.%02d_bin", $leoid, $ndump, $antenna_id);

  #
  ## Convert the output hrbin file into a correct opnGps file.
  ## This entails sorting it by PRN number and adding headers
  ## and trailers.
  #
  BJTools::sortOpnGps ($if, $outfile, {VERSION => 2});
  print "FILE CREATION: $outfile\n";

}

#/**----------------------------------------------------------------------
# @sub       champ_unwrap
#
# Subroutine to get rid of the GFZ wrappers from CHAMP data.
# These wrappers were formerly removed by 'champ_unwrap.c' from JPL, and
# consist simply of 4 bytes header per turbo binary packet, two bytes
# tag and two bytes length.
#
# @parameter  $textref -- reference to string containing file data
# @return     Nothing--$textref will be modified in-place.
# @exception  Thrown.
# ----------------------------------------------------------------------*/
sub champ_unwrap {
  my $textref = shift;
  my $textout;
  my $idx = 0;
  while (1) {
    my ($id, $len) = unpack ("S>S>", substr ($$textref, $idx, 4));
    last if (!defined($id));
    die "Bad CHAMP packet header $id at $idx" if ($id != 0xf100 && $id != 0xf200);
    $textout .= substr($$textref, $idx+4, $len);
    $idx+=$len+4;
  }
  return \$textout;
}
