#!/ops/tools/bin/perl
#
##  Copyright (c) 1995-2010 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------    
# @file       cnofsZero2one.pl
#
# Used in real-time processing.  Convert a single dump of C/NOFS
# into level 1a files (opnGps, podCrx, goxSOH).
#
# @author     Doug Hunt
# @since      03/12/2010
# @version    $URL: svn://svn1.cosmic.ucar.edu/trunk/src/MOOutils/cnofsZero2one.pl $ $Id: cnofsZero2one.pl 24678 2019-05-15 01:06:37Z vanhove $
# cdaacTask   yes
# @usage      cnofsZero2one.pl fullpath
# -----------------------------------------------------------------------*/
EOD

$|++;

#/** @call BJutils::BJTools.pm:sortNav */
use lib qw (.);
use strict;
use warnings;
use BJTools;    # BJ unpacking library
use Level1aTools;
use File::Temp qw(tempdir);
use CnofsParser;

if (@ARGV < 1) {
  print $pkgdoc;
  exit -1;
}

my $fullpath = shift;
# Find date, LEO ID and dump number from input path: grcLv0_2016.366.02.36.001_bj
my ($date, $hr, $min) = ($fullpath =~ /cnfLv0_(\d\d\d\d\.\d\d\d)\.(\d\d)\.(\d\d)/);
my $ndump    = 1; # Some logic is needed to keep dump ids straight!  We'll just call it '1' for now...
my $leoid    = 1;

print "Processing $fullpath...\n";


# Create a temporary working directory.  This returns a ref to a scalar, 
# the scalar contains the name of the temp dir.
# Create a temporary working directory.  This returns a ref to a scalar,
# the scalar contains the name of the temp dir.
my $tmpdir = tempdir( CLEANUP => 1 );
my $tmpdirref = \$tmpdir;

my $parser = CnofsParser->new({TMPDIR => $tmpdirref});

# Create low rate, high rate and NAV intermediate data files in $$tmpdir for all input binary files
# system "rm -f /tmp/grace_residuals*"; # DEBUG
$parser->parseLv0($fullpath);

$parser->report; # print report of parsing success

my ($filestart, $fileend) = ($parser->{FIRSTTIME}, $parser->{LASTTIME});

#
## Convert the high rate data into CDAAC opnGps format
#
my @hrfiles  = glob ("$tmpdir/hirate*");
foreach my $hrfile (@hrfiles) {

  my ($antenna_id) = $hrfile =~ /hirate\.bin(\d+)/;
  my $outfile = sprintf ("opnGps_$date.%03d.%02d.%02d_bin", $leoid, $ndump, $antenna_id);

  print "Spline interpolating 1Hz L2 phase to 50Hz...\n";
  CnofsParser::fill_onesec_l2 ($hrfile, "$hrfile.fill");

  print "Converting high rate data file to opnGps: $outfile\n";
  Level1aTools::sortOpnGps ("$hrfile.fill", $outfile, {VERSION => 2});

  # Notify Scheduler via shared queue
  print ("FILE CREATION: $outfile\n");
}

#--------------------------------------------------------------------------------------

my @lrfiles = glob ("$tmpdir/lorate*");

print "Converting ASCII low rate data to rinex format (10 second data)\n";
foreach my $lrtmp (@lrfiles) {

  # sort by time
  Level1aTools::sortLrObs($lrtmp, $lrtmp);

  my ($antenna_id) = $lrtmp =~ /lorate\.txt(\d+)/;
  my $outfile = sprintf ("podCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  my $leoname = 'CNFS';
  my $mission = 'cnofs';
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 10, {compress => 0});
  unlink $outfile if (-s $outfile < 100);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

  $outfile = sprintf ("ionCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 1, {compress => 0});
  unlink $outfile if (-s $outfile < 100);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

}

my @navfiles = glob ("$tmpdir/navg*");

print "Sorting NAV files\n";
foreach my $navtmp (@navfiles) {

  my $outfile = sprintf ("goxSOH_$date.%03d.%02d_txt", $leoid, $ndump);
  BJTools::sortNav($navtmp, $outfile);

  print ("FILE CREATION: $outfile\n");
}

print "Done!!!\n";
