#!/usr/bin/env perl
#
##  Copyright (c) 1995-2011 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------
# @file       cosmicGoxZero2one.pl
#
# Convert one goxBin file into standard level 1a files:
#
# podGps -- Low rate Binex GPS data
# opnGps -- High rate occultation and reference link data (custom binary format)
# goxSOH -- ASCII State of Health data
# scnRaw -- Scintillation data
#
# For details on these formats, see:
#
# http://cosmic.cosmic.ucar.edu/cdaac/fileFormats/
#
# This is a simplified code which does not deal with COSMIC attitude data
# or with other CDAAC-specific complications such as the database or the
# CDAAC directory tree layout.  The primary purpose for this code is
# for documentation of the COSMIC level 0 data format to assist other
# data centers with processing COSMIC GOX data.
#
# @author     Doug Hunt
# @since      06/23/2011
# @version    $URL$ $Id$
# @cdaacTask  no
# @usage      cosmicGoxZero2one.pl goxBin_2011.134.005.02.B_bj
# @           The input file is specified as a full path or the simple file
# @           name of a file in the current directory.  Output files
# @           are placed in the current directory.
# @           If --l2c is specified, use L2C data if available.
# @           To turn off the L2C data format, use --nol2c.
# @           This will reduce the size of the output opnGps files.
# -----------------------------------------------------------------------*/
EOD

$|++;

# Perl pragmas
use lib qw (.);
use strict;
use warnings;

# Perl standard libraries
use File::Temp qw(tempdir);
use Getopt::Long;

# Perl extensions
use PDL;

# CDAAC libraries
use BJTools;
use CosmicParser;
use Level1aTools;

if (@ARGV < 1) {
  print $pkgdoc;
  exit -1;
}

my $inpath  = shift;

# Deal with command line options
my $l2c    = 1;
GetOptions (
            "l2c!"          => \$l2c,
	    ) || die "Cannot parse command line options!";

print "cosmicZero2one.pl:  processing $inpath\n";

# Find date, LEO ID and dump number from input path: goxBin_2011.134.005.02.B_bj
my ($date, $leoid, $ndump, $gox_side) = ($inpath =~ /goxBin_(\d\d\d\d\.\d\d\d)\.(\d\d\d)\.(\d\d)\.(A|B)_bj/);

# Create a temporary working directory.  This returns a ref to a scalar,
# the scalar contains the name of the temp dir.
my $tmpdir = tempdir( CLEANUP => 1 );
my $tmpdirref = \$tmpdir;

# start fresh!
unlink glob ("$tmpdir/*");

print "Processing $inpath...\n";

my $hrtype = $l2c ? 3 : 1; # 3 means 'l2c open loop' 1 means 'normal open loop'
my $parser = CosmicParser->new({TMPDIR => $tmpdirref, HRTYPE => $hrtype});
$parser->parseLv0($inpath);
$parser->report; # print report of parsing success

print "Extracting LOG data\n";
my $logdata = '';
my @data_types = grep { -s "$tmpdir/$_" } (qw(logm.txt cmdr.txt fdir.txt trkd.txt opar.txt antc.txt navg.txt));
$logdata .= BJTools::logData ("$tmpdir/$_") foreach (@data_types);

# Print out any GOX log data
if (length($logdata) > 0) {
  my $logfile = sprintf ("goxSOH_$date.%03d.%02d_txt", $leoid, $ndump);
  open (OUT, ">$logfile") || die "Cannot open $logfile for output";
  print OUT $logdata;
  close OUT;
  print "FILE CREATION: $logfile\n";
}

# Extract any scintillation data
print "Extracting SCIN data\n";

# %scindata is a hash by antenna id:
# $scindata{antid}{time}{prn}[CAsnr, SCIN]
my %scindata = BJTools::scinData ("$tmpdir/scin.txt");

foreach my $antid (sort keys %scindata) {

  my $scinfile = sprintf ("scnRaw_$date.%03d.%02d.%02d_txt", $leoid, $ndump, $antid);
  open  (my $OUT, '>', $scinfile) or die "Cannot open $scinfile for output";
  print {$OUT} "# GPS_seconds Number_of_GPS_satellites_this_epoch list_of_PRNs\n";
  print {$OUT} "# One PRN per line:  CA_SNR SCIN_value\n";
  foreach my $time (sort keys %{$scindata{$antid}}) {
    my @prns = sort keys %{$scindata{$antid}{$time}};
    printf {$OUT} "%10d  %2d".(' %2d' x @prns)."\n", $time, scalar(@prns), @prns;
    foreach my $prn (@prns) {
      printf {$OUT} "%10d%10d\n", @{$scindata{$antid}{$time}{$prn}};
    }
  }
  close $OUT;
  print "FILE CREATION: $scinfile\n";
}

#
## Separate the lorate.txt file into separate files per antenna ID:
## lorate.txtTT, where TT is the antenna ID
#

# Now do merging of tracks when splitting. D. Hunt 3/10/2009
my @files = (-s "$tmpdir/lorate.txt") ? BJTools::splitLRdata_merge ("$tmpdir/lorate.txt")
          : ();

print "Converting ASCII output of BJfmtl to BINEX 0x7f_00 format\n";

foreach my $lrtmp (@files) {

  my ($antenna_id) = $lrtmp =~ /lorate\.txt(\d+)/;

  my $lrfile = sprintf ("podCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);

  # Call blackJack2rinex with 'compress => 0' option so we do not have to mess with rinex compression
  Level1aTools::blackJack2rinex($lrtmp,$lrfile,$antenna_id,sprintf("C%03d",$leoid),'cosmic',1,{compress => 0});

  do { unlink $lrfile; next } if (-s $lrfile < 1000);  # skip small (header only) files
  print "FILE CREATION: $lrfile\n";

}

print "Converting high rate data to opnGps format\n";
my @if  = glob ("$tmpdir/hirate*");
foreach my $if (sort @if) {

  my ($antenna_id) = $if =~ /hirate\.bin(\d+)/;

  my $outfile = sprintf ("opnGps_$date.%03d.%02d.%02d_bin", $leoid, $ndump, $antenna_id);

  #
  ## Convert the output hrbin file into a correct opnGps file.
  ## This entails sorting it by PRN number and adding headers
  ## and trailers.
  #
  my $version = $l2c ? 3 : 1; # either open loop CA or open loop l2c
  BJTools::sortOpnGps ($if, $outfile, {VERSION => $version});
  print "FILE CREATION: $outfile\n";

}

print "Done!\n";

exit;


