#!/usr/bin/env perl
#
##  Copyright (c) 1995-2017 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------    
# @file       graceZero2one.pl
# Convert one level 0 GRACE data file into level 1 files as specified
# in the COSMIC level 1 data standard:
#
# http://cdaac-www.cosmic.ucar.edu/cdaac/fileFormats/
#
# @author     Doug Hunt
# @since      3/31/2017
# @version    $URL$ $Id$
# @cdaacTask  yes
# @usage      graceZero2one.pl grcLv0_2016.366.02.36.001_bj
# -----------------------------------------------------------------------*/
EOD

$|++;

# Perl pragmas
use lib qw (.);
use strict;
use warnings;

# Perl standard libraries
use File::Temp qw(tempdir);
use Getopt::Long;

# Perl extensions
use PDL;

# CDAAC libraries
use BJTools;
use GraceParser;
use Level1aTools;

if (@ARGV < 1) {
  print $pkgdoc;
  exit -1;
}

my $inpath = shift;

print "graceZero2one.pl:  processing $inpath\n";

# Find date, LEO ID and dump number from input path:
# Note:
# The temporary switching of the RO experiment between GRACE-A(1) and GRACE-B(2)
# since 2014.208 lead to the following update in the filenameing convention at
# CDAAC effective with 2014.090:
#  - legacy: grcLv0_2014.089.00.53_bj (only GRACE-A)
#  - current: grcLv0_2014.090.00.09.001_bj (GRACE-A)
#             grcLv0_2014.208.07.58.002_bj (GRACE-B)
my $file_pattern_legacy = qr/grcLv0_(\d\d\d\d\.\d\d\d)\.(\d\d)\.(\d\d)_bj/;
my $file_pattern = qr/grcLv0_(\d\d\d\d\.\d\d\d)\.(\d\d)\.(\d\d)\.(\d\d\d)_bj/;
my ($date, $hr, $min, $leoid);
if ($inpath =~ $file_pattern_legacy) {
  ($date, $hr, $min) = ($inpath =~ $file_pattern_legacy);
  $leoid = 1; # Not given in legacy filenamening
}
elsif ($inpath =~ $file_pattern) {
  ($date, $hr, $min, $leoid) = ($inpath =~ $file_pattern);
}
else {
  die "ERROR: Unrecognized file pattern '$inpath'.\n";
}
my $ndump = 1; # Some logic is needed to keep dump ids straight!  We'll just call it '1' for now...

# Create a temporary working directory.  This returns a ref to a scalar,
# the scalar contains the name of the temp dir.
my $tmpdir = tempdir( CLEANUP => 1 );
my $tmpdirref = \$tmpdir;

# start fresh!
unlink glob ("$tmpdir/*");

my $parser = GraceParser->new({TMPDIR => $tmpdirref});
$parser->parseLv0($inpath);
$parser->report; # print report of parsing success

print "Extracting NAV data\n";
my $logdata = BJTools::logData ("$tmpdir/navg.txt");

# Print out any GOX log data
if (length($logdata) > 0) {
  my $logfile = sprintf ("goxSOH_$date.%03d.%02d_txt", $leoid, $ndump);
  open (OUT, ">$logfile") || die "Cannot open $logfile for output";
  print OUT $logdata;
  close OUT;
  print "FILE CREATION: $logfile\n";
}

my @lrfiles = glob ("$tmpdir/lorate*");

print "Converting ASCII low rate data to level 1a format (RINEX)\n";
foreach my $lrtmp (@lrfiles) {

  # sort by time
  Level1aTools::sortLrObs($lrtmp, $lrtmp);

  my ($antenna_id) = $lrtmp =~ /lorate\.txt(\d+)/;

  #
  ## Write out 10 second data in compressed RINEX format
  #
  my $outfile = sprintf ("podCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  my $leoname = $leoid == 1 ? 'GRC1' : 'GRC2';
  my $mission = 'grace';
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 10, {compress => 0});
  unlink $outfile if (-s $outfile && -s $outfile < 2000);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

  #
  ## Write out 1 second data in compressed RINEX format
  #
  $outfile = sprintf ("ionCrx_$date.%03d.%02d.%02d_rnx", $leoid, $ndump, $antenna_id);
  Level1aTools::blackJack2rinex ($lrtmp, $outfile, $antenna_id, $leoname, $mission, 1, {compress => 0});
  unlink $outfile if (-s $outfile && -s $outfile < 2000);  # skip small (header only) files
  print ("FILE CREATION: $outfile\n") if (-s $outfile);

}

print "Converting high rate data to opnGps format\n";
my @if  = glob ("$tmpdir/hirate*");
foreach my $if (sort @if) {

  my ($antenna_id) = $if =~ /hirate\.bin(\d+)/;

  my $outfile = sprintf ("opnGps_$date.%03d.%02d.%02d_bin", $leoid, $ndump, $antenna_id);

  #
  ## Convert the output hrbin file into a correct opnGps file.
  ## This entails sorting it by PRN number and adding headers
  ## and trailers.
  #
  BJTools::sortOpnGps ($if, $outfile, {VERSION => 2});
  print "FILE CREATION: $outfile\n";

}

print "Done!\n";

exit;

